#pragma once 
// IWYU pragma: private <quark.h>

#include <fmt/base.h>
#include <fmt/ostream.h>
#include <type_traits>

namespace qk {
	enum class EventType {
		None = 0,

		WindowClose,
		WindowResize,
		WindowFocus,
		WindowLostFocus,
		WindowMoved,

		KeyPressed,
		KeyReleased,

		MouseButtonPressed,
		MouseButtonReleased,
		MouseMoved,
		MouseScrolled
	};

	enum EventCategory {
		None = 0,

		EventCategoryApplication = 0x1,
		EventCategoryInput       = 0x2,
		EventCategoryKeyboard    = 0x4,
		EventCategoryMouse       = 0x8,
		EventCategoryMouseButton = 0x10,
	};

#define EVENT_CLASS_TYPE(type) \
	static EventType GetStaticType() { return (EventType::type); } \
	virtual EventType GetEventType() const override { return GetStaticType(); } \
	virtual const char* GetName() const override { return #type; }

#define EVENT_CLASS_CATEGORY(category) \
	virtual int GetCategoryFlags() const override { return category; }

	class Event {
		friend class EventDispatcher;

	public:
		virtual EventType GetEventType() const = 0;
		virtual const char* GetName() const = 0;
		virtual int GetCategoryFlags() const = 0;

		virtual operator std::string() const {
			return GetName();
		}

		inline bool IsInCagegory(EventCategory category) const {
			return (GetCategoryFlags() & category) != 0;
		}

	protected:
		bool handled = false;
	};

	class EventDispatcher {
		template <typename T>
		using EventFn = std::function<bool(T&)>;

	public:
		EventDispatcher(Event& event) :
			event(event)
		{ }

		template <typename T>
		requires std::is_base_of_v<Event, T>
		bool Dispatch(EventFn<T> func) {
			if (event.GetEventType() == T::GetStaticType()) {
				event.handled = func(*(T*)&event);
				return true;
			}

			return false;
		}

	private:
		Event& event;
	};

	inline std::ostream& operator<<(std::ostream& os, const Event& event) {
		os << std::string(event);
		return os;
	}

}

template <typename T>
struct fmt::formatter<T, std::enable_if_t<std::is_base_of_v<qk::Event, T>, char>> :
fmt::ostream_formatter {};
